@@ -2,12 +2,20 @@
  * load methods from eval.c
  */
 
-#include "ruby/encoding.h"
-#include "ruby/util.h"
-#include "internal.h"
 #include "dln.h"
 #include "eval_intern.h"
+#include "internal.h"
+#include "internal/dir.h"
+#include "internal/error.h"
+#include "internal/file.h"
+#include "internal/load.h"
+#include "internal/parse.h"
+#include "internal/thread.h"
+#include "internal/variable.h"
+#include "iseq.h"
 #include "probes.h"
+#include "ruby/encoding.h"
+#include "ruby/util.h"
 
 static VALUE ruby_dln_librefs;
 
@@ -27,13 +35,6 @@ static const char *const loadable_ext[] = {
     0
 };
 
-VALUE
-rb_get_load_path(void)
-{
-    VALUE load_path = GET_VM()->load_path;
-    return load_path;
-}
-
 enum expand_type {
     EXPAND_ALL,
     EXPAND_RELATIVE,
@@ -54,7 +55,6 @@ rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *h
     VALUE expanded_load_path = vm->expanded_load_path;
     VALUE ary;
     long i;
-    int level = rb_safe_level();
 
     ary = rb_ary_tmp_new(RARRAY_LEN(load_path));
     for (i = 0; i < RARRAY_LEN(load_path); ++i) {
@@ -64,7 +64,7 @@ rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *h
 	as_str = path = RARRAY_AREF(load_path, i);
 	is_string = RB_TYPE_P(path, T_STRING) ? 1 : 0;
 	non_cache = !is_string ? 1 : 0;
-	as_str = rb_get_path_check_to_string(path, level);
+        as_str = rb_get_path_check_to_string(path);
 	as_cstr = RSTRING_PTR(as_str);
 
 	if (!non_cache) {
@@ -85,8 +85,8 @@ rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *h
 	/* Freeze only string object. We expand other objects every time. */
 	if (is_string)
 	    rb_str_freeze(path);
-	as_str = rb_get_path_check_convert(path, as_str, level);
-	expanded_path = rb_check_realpath(Qnil, as_str);
+        as_str = rb_get_path_check_convert(as_str);
+	expanded_path = rb_check_realpath(Qnil, as_str, NULL);
 	if (NIL_P(expanded_path)) expanded_path = as_str;
 	rb_ary_push(ary, rb_fstring(expanded_path));
     }
@@ -142,8 +142,9 @@ rb_get_expanded_load_path(void)
 }
 
 static VALUE
-load_path_getter(ID id, rb_vm_t *vm)
+load_path_getter(ID id, VALUE * p)
 {
+    rb_vm_t *vm = (void *)p;
     return vm->load_path;
 }
 
@@ -153,6 +154,12 @@ get_loaded_features(void)
     return GET_VM()->loaded_features;
 }
 
+static VALUE
+get_LOADED_FEATURES(ID _x, VALUE *_y)
+{
+    return get_loaded_features();
+}
+
 static void
 reset_loaded_features_snapshot(void)
 {
@@ -473,9 +480,7 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
     }
     if (st_get_key(loading_tbl, (st_data_t)feature, &data)) {
 	if (fn) *fn = (const char*)data;
-      loading:
-	if (!ext) return 'u';
-	return !IS_RBEXT(ext) ? 's' : 'r';
+        goto loading;
     }
     else {
 	VALUE bufstr;
@@ -507,6 +512,10 @@ rb_feature_p(const char *feature, const char *ext, int rb, int expanded, const c
 	rb_str_resize(bufstr, 0);
     }
     return 0;
+
+  loading:
+    if (!ext) return 'u';
+    return !IS_RBEXT(ext) ? 's' : 'r';
 }
 
 int
@@ -566,10 +575,27 @@ rb_provide(const char *feature)
 }
 
 NORETURN(static void load_failed(VALUE));
-const rb_iseq_t *rb_iseq_load_iseq(VALUE fname);
 
-static int
-rb_load_internal0(rb_execution_context_t *ec, VALUE fname, int wrap)
+static inline void
+load_iseq_eval(rb_execution_context_t *ec, VALUE fname)
+{
+    const rb_iseq_t *iseq = rb_iseq_load_iseq(fname);
+
+    if (!iseq) {
+        rb_ast_t *ast;
+        VALUE parser = rb_parser_new();
+        rb_parser_set_context(parser, NULL, FALSE);
+        ast = (rb_ast_t *)rb_parser_load_file(parser, fname);
+        iseq = rb_iseq_new_top(&ast->body, rb_fstring_lit("<top (required)>"),
+                               fname, rb_realpath_internal(Qnil, fname, 1), NULL);
+        rb_ast_dispose(ast);
+    }
+    rb_exec_event_hook_script_compiled(ec, iseq, Qnil);
+    rb_iseq_eval(iseq);
+}
+
+static inline enum ruby_tag_type
+load_wrapping(rb_execution_context_t *ec, VALUE fname)
 {
     enum ruby_tag_type state;
     rb_thread_t *th = rb_ec_thread_ptr(ec);
@@ -579,36 +605,17 @@ rb_load_internal0(rb_execution_context_t *ec, VALUE fname, int wrap)
     rb_thread_t *volatile th0 = th;
 #endif
 
-    th->ec->errinfo = Qnil; /* ensure */
+    ec->errinfo = Qnil; /* ensure */
 
-    if (!wrap) {
-	th->top_wrapper = 0;
-    }
-    else {
-	/* load in anonymous module as toplevel */
-	th->top_self = rb_obj_clone(rb_vm_top_self());
-	th->top_wrapper = rb_module_new();
-	rb_extend_object(th->top_self, th->top_wrapper);
-    }
+    /* load in anonymous module as toplevel */
+    th->top_self = rb_obj_clone(rb_vm_top_self());
+    th->top_wrapper = rb_module_new();
+    rb_extend_object(th->top_self, th->top_wrapper);
 
-    EC_PUSH_TAG(th->ec);
+    EC_PUSH_TAG(ec);
     state = EC_EXEC_TAG();
     if (state == TAG_NONE) {
-	rb_ast_t *ast;
-	const rb_iseq_t *iseq;
-
-	if ((iseq = rb_iseq_load_iseq(fname)) != NULL) {
-	    /* OK */
-	}
-	else {
-	    VALUE parser = rb_parser_new();
-	    rb_parser_set_context(parser, NULL, FALSE);
-	    ast = (rb_ast_t *)rb_parser_load_file(parser, fname);
-	    iseq = rb_iseq_new_top(&ast->body, rb_fstring_cstr("<top (required)>"),
-			    fname, rb_realpath_internal(Qnil, fname, 1), NULL);
-	    rb_ast_dispose(ast);
-	}
-	rb_iseq_eval(iseq);
+        load_iseq_eval(ec, fname);
     }
     EC_POP_TAG();
 
@@ -618,61 +625,54 @@ rb_load_internal0(rb_execution_context_t *ec, VALUE fname, int wrap)
 #endif
     th->top_self = self;
     th->top_wrapper = wrapper;
-
-    if (state) {
-	/* usually state == TAG_RAISE only, except for
-	 * rb_iseq_load_iseq case */
-	VALUE exc = rb_vm_make_jump_tag_but_local_jump(state, Qundef);
-	if (NIL_P(exc)) return state;
-	th->ec->errinfo = exc;
-	return TAG_RAISE;
-    }
-
-    if (!NIL_P(th->ec->errinfo)) {
-	/* exception during load */
-	return TAG_RAISE;
-    }
     return state;
 }
 
+static inline void
+raise_load_if_failed(rb_execution_context_t *ec, enum ruby_tag_type state)
+{
+    if (state) {
+        rb_vm_jump_tag_but_local_jump(state);
+    }
+
+    if (!NIL_P(ec->errinfo)) {
+        rb_exc_raise(ec->errinfo);
+    }
+}
+
 static void
 rb_load_internal(VALUE fname, int wrap)
 {
     rb_execution_context_t *ec = GET_EC();
-    int state = rb_load_internal0(ec, fname, wrap);
-    if (state) {
-	if (state == TAG_RAISE) rb_exc_raise(ec->errinfo);
-	EC_JUMP_TAG(ec, state);
+    enum ruby_tag_type state = TAG_NONE;
+    if (wrap) {
+        state = load_wrapping(ec, fname);
     }
-}
-
-static VALUE
-file_to_load(VALUE fname)
-{
-    VALUE tmp = rb_find_file(FilePathValue(fname));
-    if (!tmp) load_failed(fname);
-    return tmp;
+    else {
+        load_iseq_eval(ec, fname);
+    }
+    raise_load_if_failed(ec, state);
 }
 
 void
 rb_load(VALUE fname, int wrap)
 {
-    rb_load_internal(file_to_load(fname), wrap);
+    VALUE tmp = rb_find_file(FilePathValue(fname));
+    if (!tmp) load_failed(fname);
+    rb_load_internal(tmp, wrap);
 }
 
 void
 rb_load_protect(VALUE fname, int wrap, int *pstate)
 {
     enum ruby_tag_type state;
-    volatile VALUE path = 0;
 
     EC_PUSH_TAG(GET_EC());
     if ((state = EC_EXEC_TAG()) == TAG_NONE) {
-	path = file_to_load(fname);
+        rb_load(fname, wrap);
     }
     EC_POP_TAG();
 
-    if (state == TAG_NONE) state = rb_load_internal0(GET_EC(), path, wrap);
     if (state != TAG_NONE) *pstate = state;
 }
 
@@ -680,24 +680,40 @@ rb_load_protect(VALUE fname, int wrap, int *pstate)
  *  call-seq:
  *     load(filename, wrap=false)   -> true
  *
- *  Loads and executes the Ruby
- *  program in the file _filename_. If the filename does not
- *  resolve to an absolute path, the file is searched for in the library
- *  directories listed in <code>$:</code>. If the optional _wrap_
- *  parameter is +true+, the loaded script will be executed
- *  under an anonymous module, protecting the calling program's global
- *  namespace. In no circumstance will any local variables in the loaded
- *  file be propagated to the loading environment.
+ *  Loads and executes the Ruby program in the file _filename_.
+ *
+ *  If the filename is an absolute path (e.g. starts with '/'), the file
+ *  will be loaded directly using the absolute path.
+ *
+ *  If the filename is an explicit relative path (e.g. starts with './' or
+ *  '../'), the file will be loaded using the relative path from the current
+ *  directory.
+ *
+ *  Otherwise, the file will be searched for in the library
+ *  directories listed in <code>$LOAD_PATH</code> (<code>$:</code>).
+ *  If the file is found in a directory, it will attempt to load the file
+ *  relative to that directory.  If the file is not found in any of the
+ *  directories in <code>$LOAD_PATH</code>, the file will be loaded using
+ *  the relative path from the current directory.
+ *
+ *  If the file doesn't exist when there is an attempt to load it, a
+ *  LoadError will be raised.
+ *
+ *  If the optional _wrap_ parameter is +true+, the loaded script will
+ *  be executed under an anonymous module, protecting the calling
+ *  program's global namespace. In no circumstance will any local
+ *  variables in the loaded file be propagated to the loading
+ *  environment.
  */
 
 static VALUE
-rb_f_load(int argc, VALUE *argv)
+rb_f_load(int argc, VALUE *argv, VALUE _)
 {
     VALUE fname, wrap, path, orig_fname;
 
     rb_scan_args(argc, argv, "11", &fname, &wrap);
 
-    orig_fname = rb_get_path_check_to_string(fname, rb_safe_level());
+    orig_fname = rb_get_path_check_to_string(fname);
     fname = rb_str_encode_ospath(orig_fname);
     RUBY_DTRACE_HOOK(LOAD_ENTRY, RSTRING_PTR(orig_fname));
 
@@ -729,7 +745,7 @@ load_lock(const char *ftptr)
     }
     else if (imemo_type_p(data, imemo_memo)) {
 	struct MEMO *memo = MEMO_CAST(data);
-	void (*init)(void) = (void (*)(void))memo->u3.func;
+        void (*init)(void) = memo->u3.func;
 	data = (st_data_t)rb_thread_shield_new();
 	st_insert(loading_tbl, (st_data_t)ftptr, data);
 	(*init)();
@@ -785,8 +801,10 @@ load_unlock(const char *ftptr, int done)
  *  Loads the given +name+, returning +true+ if successful and +false+ if the
  *  feature is already loaded.
  *
- *  If the filename does not resolve to an absolute path, it will be searched
- *  for in the directories listed in <code>$LOAD_PATH</code> (<code>$:</code>).
+ *  If the filename neither resolves to an absolute path nor starts with
+ *  './' or '../', the file will be searched for in the library
+ *  directories listed in <code>$LOAD_PATH</code> (<code>$:</code>).
+ *  If the filename starts with './' or '../', resolution is based on Dir.pwd.
  *
  *  If the filename has the extension ".rb", it is loaded as a source file; if
  *  the extension is ".so", ".o", or ".dll", or the default shared library
@@ -817,7 +835,7 @@ load_unlock(const char *ftptr, int done)
 VALUE
 rb_f_require(VALUE obj, VALUE fname)
 {
-    return rb_require_safe(fname, rb_safe_level());
+    return rb_require_string(fname);
 }
 
 /*
@@ -836,11 +854,13 @@ rb_f_require_relative(VALUE obj, VALUE fname)
 	rb_loaderror("cannot infer basepath");
     }
     base = rb_file_dirname(base);
-    return rb_require_safe(rb_file_absolute_path(fname, base), rb_safe_level());
+    return rb_require_string(rb_file_absolute_path(fname, base));
 }
 
+typedef int (*feature_func)(const char *feature, const char *ext, int rb, int expanded, const char **fn);
+
 static int
-search_required(VALUE fname, volatile VALUE *path, int safe_level)
+search_required(VALUE fname, volatile VALUE *path, feature_func rb_feature_p)
 {
     VALUE tmp;
     char *ext, *ftptr;
@@ -855,7 +875,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 'r';
 	    }
-	    if ((tmp = rb_find_file_safe(fname, safe_level)) != 0) {
+            if ((tmp = rb_find_file(fname)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
 		if (!rb_feature_p(ftptr, ext, TRUE, TRUE, &loading) || loading)
 		    *path = tmp;
@@ -871,7 +891,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	    tmp = rb_str_subseq(fname, 0, ext - RSTRING_PTR(fname));
 #ifdef DLEXT2
 	    OBJ_FREEZE(tmp);
-	    if (rb_find_file_ext_safe(&tmp, loadable_ext + 1, safe_level)) {
+            if (rb_find_file_ext(&tmp, loadable_ext + 1)) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
 		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
 		    *path = tmp;
@@ -880,7 +900,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 #else
 	    rb_str_cat2(tmp, DLEXT);
 	    OBJ_FREEZE(tmp);
-	    if ((tmp = rb_find_file_safe(tmp, safe_level)) != 0) {
+            if ((tmp = rb_find_file(tmp)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
 		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
 		    *path = tmp;
@@ -893,7 +913,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 		if (loading) *path = rb_filesystem_str_new_cstr(loading);
 		return 's';
 	    }
-	    if ((tmp = rb_find_file_safe(fname, safe_level)) != 0) {
+            if ((tmp = rb_find_file(fname)) != 0) {
 		ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
 		if (!rb_feature_p(ftptr, ext, FALSE, TRUE, &loading) || loading)
 		    *path = tmp;
@@ -906,7 +926,7 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	return 'r';
     }
     tmp = fname;
-    type = rb_find_file_ext_safe(&tmp, loadable_ext, safe_level);
+    type = rb_find_file_ext(&tmp, loadable_ext);
     switch (type) {
       case 0:
 	if (ft)
@@ -916,10 +936,9 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 
       default:
 	if (ft) {
-	  statically_linked:
-	    if (loading) *path = rb_filesystem_str_new_cstr(loading);
-	    return ft;
+            goto statically_linked;
 	}
+        /* fall through */
       case 1:
 	ext = strrchr(ftptr = RSTRING_PTR(tmp), '.');
 	if (rb_feature_p(ftptr, ext, !--type, TRUE, &loading) && !loading)
@@ -927,6 +946,10 @@ search_required(VALUE fname, volatile VALUE *path, int safe_level)
 	*path = tmp;
     }
     return type ? 's' : 'r';
+
+  statically_linked:
+    if (loading) *path = rb_filesystem_str_new_cstr(loading);
+    return ft;
 }
 
 static void
@@ -942,6 +965,57 @@ load_ext(VALUE path)
     return (VALUE)dln_load(RSTRING_PTR(path));
 }
 
+static int
+no_feature_p(const char *feature, const char *ext, int rb, int expanded, const char **fn)
+{
+    return 0;
+}
+
+// Documented in doc/globals.rdoc
+VALUE
+rb_resolve_feature_path(VALUE klass, VALUE fname)
+{
+    VALUE path;
+    int found;
+    VALUE sym;
+
+    fname = rb_get_path(fname);
+    path = rb_str_encode_ospath(fname);
+    found = search_required(path, &path, no_feature_p);
+
+    switch (found) {
+      case 'r':
+        sym = ID2SYM(rb_intern("rb"));
+        break;
+      case 's':
+        sym = ID2SYM(rb_intern("so"));
+        break;
+      default:
+        load_failed(fname);
+    }
+
+    return rb_ary_new_from_args(2, sym, path);
+}
+
+static void
+ext_config_push(rb_thread_t *th, struct rb_ext_config *prev)
+{
+    *prev = th->ext_config;
+    th->ext_config = (struct rb_ext_config){0};
+}
+
+static void
+ext_config_pop(rb_thread_t *th, struct rb_ext_config *prev)
+{
+    th->ext_config = *prev;
+}
+
+void
+rb_ext_ractor_safe(bool flag)
+{
+    GET_THREAD()->ext_config.ractor_safe = flag;
+}
+
 /*
  * returns
  *  0: if already loaded (false)
@@ -949,72 +1023,93 @@ load_ext(VALUE path)
  * <0: not found (LoadError)
  * >1: exception
  */
-int
-rb_require_internal(VALUE fname, int safe)
+static int
+require_internal(rb_execution_context_t *ec, VALUE fname, int exception)
 {
     volatile int result = -1;
-    rb_execution_context_t *ec = GET_EC();
+    rb_thread_t *th = rb_ec_thread_ptr(ec);
+    volatile VALUE wrapper = th->top_wrapper;
+    volatile VALUE self = th->top_self;
     volatile VALUE errinfo = ec->errinfo;
     enum ruby_tag_type state;
-    struct {
-	int safe;
-    } volatile saved;
     char *volatile ftptr = 0;
     VALUE path;
+    volatile bool reset_ext_config = false;
+    struct rb_ext_config prev_ext_config;
 
-    fname = rb_get_path_check(fname, safe);
+    fname = rb_get_path(fname);
     path = rb_str_encode_ospath(fname);
     RUBY_DTRACE_HOOK(REQUIRE_ENTRY, RSTRING_PTR(fname));
 
     EC_PUSH_TAG(ec);
-    saved.safe = rb_safe_level();
+    ec->errinfo = Qnil; /* ensure */
+    th->top_wrapper = 0;
     if ((state = EC_EXEC_TAG()) == TAG_NONE) {
 	long handle;
 	int found;
 
-	rb_set_safe_level_force(0);
-
 	RUBY_DTRACE_HOOK(FIND_REQUIRE_ENTRY, RSTRING_PTR(fname));
-	found = search_required(path, &path, safe);
+        found = search_required(path, &path, rb_feature_p);
 	RUBY_DTRACE_HOOK(FIND_REQUIRE_RETURN, RSTRING_PTR(fname));
 
 	if (found) {
-	    if (!path || !(ftptr = load_lock(RSTRING_PTR(path)))) {
+            if (!path || !(ftptr = load_lock(RSTRING_PTR(path)))) {
 		result = 0;
 	    }
 	    else if (!*ftptr) {
-		rb_provide_feature(path);
 		result = TAG_RETURN;
 	    }
 	    else {
 		switch (found) {
 		  case 'r':
-		    state = rb_load_internal0(ec, path, 0);
+                    load_iseq_eval(ec, path);
 		    break;
 
 		  case 's':
+                    reset_ext_config = true;
+                    ext_config_push(th, &prev_ext_config);
 		    handle = (long)rb_vm_call_cfunc(rb_vm_top_self(), load_ext,
 						    path, VM_BLOCK_HANDLER_NONE, path);
 		    rb_ary_push(ruby_dln_librefs, LONG2NUM(handle));
 		    break;
 		}
-		if (!state) {
-		    rb_provide_feature(path);
-		    result = TAG_RETURN;
-		}
+                result = TAG_RETURN;
 	    }
 	}
     }
     EC_POP_TAG();
-    load_unlock(ftptr, !state);
 
-    rb_set_safe_level_force(saved.safe);
+    th = rb_ec_thread_ptr(ec);
+    th->top_self = self;
+    th->top_wrapper = wrapper;
+    if (reset_ext_config) ext_config_pop(th, &prev_ext_config);
+
+    if (ftptr) load_unlock(RSTRING_PTR(path), !state);
+
     if (state) {
+        if (state == TAG_FATAL) {
+            EC_JUMP_TAG(ec, state);
+        }
+        else if (exception) {
+            /* usually state == TAG_RAISE only, except for
+             * rb_iseq_load_iseq in load_iseq_eval case */
+            VALUE exc = rb_vm_make_jump_tag_but_local_jump(state, Qundef);
+            if (!NIL_P(exc)) ec->errinfo = exc;
+            return TAG_RAISE;
+        }
+        else if (state == TAG_RETURN) {
+            return TAG_RAISE;
+        }
 	RB_GC_GUARD(fname);
 	/* never TAG_RETURN */
 	return state;
     }
+    if (!NIL_P(ec->errinfo)) {
+        if (!exception) return TAG_RAISE;
+        rb_exc_raise(ec->errinfo);
+    }
 
+    if (result == TAG_RETURN) rb_provide_feature(path);
     ec->errinfo = errinfo;
 
     RUBY_DTRACE_HOOK(REQUIRE_RETURN, RSTRING_PTR(fname));
@@ -1022,24 +1117,32 @@ rb_require_internal(VALUE fname, int safe)
     return result;
 }
 
+int
+rb_require_internal(VALUE fname)
+{
+    rb_execution_context_t *ec = GET_EC();
+    return require_internal(ec, fname, 1);
+}
+
 int
 ruby_require_internal(const char *fname, unsigned int len)
 {
     struct RString fake;
     VALUE str = rb_setup_fake_str(&fake, fname, len, 0);
-    int result = rb_require_internal(str, 0);
+    rb_execution_context_t *ec = GET_EC();
+    int result = require_internal(ec, str, 0);
     rb_set_errinfo(Qnil);
     return result == TAG_RETURN ? 1 : result ? -1 : 0;
 }
 
 VALUE
-rb_require_safe(VALUE fname, int safe)
+rb_require_string(VALUE fname)
 {
-    int result = rb_require_internal(fname, safe);
+    rb_execution_context_t *ec = GET_EC();
+    int result = require_internal(ec, fname, 1);
 
     if (result > TAG_RETURN) {
-	if (result == TAG_RAISE) rb_exc_raise(rb_errinfo());
-	EC_JUMP_TAG(GET_EC(), result);
+        EC_JUMP_TAG(ec, result);
     }
     if (result < 0) {
 	load_failed(fname);
@@ -1051,8 +1154,7 @@ rb_require_safe(VALUE fname, int safe)
 VALUE
 rb_require(const char *fname)
 {
-    VALUE fn = rb_str_new_cstr(fname);
-    return rb_require_safe(fn, rb_safe_level());
+    return rb_require_string(rb_str_new_cstr(fname));
 }
 
 static int
@@ -1084,8 +1186,8 @@ ruby_init_ext(const char *name, void (*init)(void))
  *  call-seq:
  *     mod.autoload(module, filename)   -> nil
  *
- *  Registers _filename_ to be loaded (using <code>Kernel::require</code>)
- *  the first time that _module_ (which may be a <code>String</code> or
+ *  Registers _filename_ to be loaded (using Kernel::require)
+ *  the first time that _module_ (which may be a String or
  *  a symbol) is accessed in the namespace of _mod_.
  *
  *     module A
@@ -1106,33 +1208,49 @@ rb_mod_autoload(VALUE mod, VALUE sym, VALUE file)
 
 /*
  *  call-seq:
- *     mod.autoload?(name)   -> String or nil
+ *     mod.autoload?(name, inherit=true)   -> String or nil
  *
  *  Returns _filename_ to be loaded if _name_ is registered as
- *  +autoload+ in the namespace of _mod_.
+ *  +autoload+ in the namespace of _mod_ or one of its ancestors.
  *
  *     module A
  *     end
  *     A.autoload(:B, "b")
  *     A.autoload?(:B)            #=> "b"
+ *
+ *  If +inherit+ is false, the lookup only checks the autoloads in the receiver:
+ *
+ *     class A
+ *       autoload :CONST, "const.rb"
+ *     end
+ *
+ *     class B < A
+ *     end
+ *
+ *     B.autoload?(:CONST)          #=> "const.rb", found in A (ancestor)
+ *     B.autoload?(:CONST, false)   #=> nil, not found in B itself
+ *
  */
 
 static VALUE
-rb_mod_autoload_p(VALUE mod, VALUE sym)
+rb_mod_autoload_p(int argc, VALUE *argv, VALUE mod)
 {
+    int recur = (rb_check_arity(argc, 1, 2) == 1) ? TRUE : RTEST(argv[1]);
+    VALUE sym = argv[0];
+
     ID id = rb_check_id(&sym);
     if (!id) {
 	return Qnil;
     }
-    return rb_autoload_p(mod, id);
+    return rb_autoload_at_p(mod, id, recur);
 }
 
 /*
  *  call-seq:
  *     autoload(module, filename)   -> nil
  *
- *  Registers _filename_ to be loaded (using <code>Kernel::require</code>)
- *  the first time that _module_ (which may be a <code>String</code> or
+ *  Registers _filename_ to be loaded (using Kernel::require)
+ *  the first time that _module_ (which may be a String or
  *  a symbol) is accessed.
  *
  *     autoload(:MyModule, "/usr/local/lib/modules/my_module.rb")
@@ -1142,7 +1260,7 @@ static VALUE
 rb_f_autoload(VALUE obj, VALUE sym, VALUE file)
 {
     VALUE klass = rb_class_real(rb_vm_cbase());
-    if (NIL_P(klass)) {
+    if (!klass) {
 	rb_raise(rb_eTypeError, "Can not set autoload on singleton class");
     }
     return rb_mod_autoload(klass, sym, file);
@@ -1150,7 +1268,7 @@ rb_f_autoload(VALUE obj, VALUE sym, VALUE file)
 
 /*
  *  call-seq:
- *     autoload?(name)   -> String or nil
+ *     autoload?(name, inherit=true)   -> String or nil
  *
  *  Returns _filename_ to be loaded if _name_ is registered as
  *  +autoload+.
@@ -1160,35 +1278,34 @@ rb_f_autoload(VALUE obj, VALUE sym, VALUE file)
  */
 
 static VALUE
-rb_f_autoload_p(VALUE obj, VALUE sym)
+rb_f_autoload_p(int argc, VALUE *argv, VALUE obj)
 {
     /* use rb_vm_cbase() as same as rb_f_autoload. */
     VALUE klass = rb_vm_cbase();
     if (NIL_P(klass)) {
 	return Qnil;
     }
-    return rb_mod_autoload_p(klass, sym);
+    return rb_mod_autoload_p(argc, argv, klass);
 }
 
 void
 Init_load(void)
 {
-#undef rb_intern
-#define rb_intern(str) rb_intern2((str), strlen(str))
     rb_vm_t *vm = GET_VM();
     static const char var_load_path[] = "$:";
     ID id_load_path = rb_intern2(var_load_path, sizeof(var_load_path)-1);
 
     rb_define_hooked_variable(var_load_path, (VALUE*)vm, load_path_getter, rb_gvar_readonly_setter);
-    rb_alias_variable(rb_intern("$-I"), id_load_path);
-    rb_alias_variable(rb_intern("$LOAD_PATH"), id_load_path);
+    rb_alias_variable(rb_intern_const("$-I"), id_load_path);
+    rb_alias_variable(rb_intern_const("$LOAD_PATH"), id_load_path);
     vm->load_path = rb_ary_new();
     vm->expanded_load_path = rb_ary_tmp_new(0);
     vm->load_path_snapshot = rb_ary_tmp_new(0);
     vm->load_path_check_cache = 0;
+    rb_define_singleton_method(vm->load_path, "resolve_feature_path", rb_resolve_feature_path, 1);
 
-    rb_define_virtual_variable("$\"", get_loaded_features, 0);
-    rb_define_virtual_variable("$LOADED_FEATURES", get_loaded_features, 0);
+    rb_define_virtual_variable("$\"", get_LOADED_FEATURES, 0);
+    rb_define_virtual_variable("$LOADED_FEATURES", get_LOADED_FEATURES, 0);
     vm->loaded_features = rb_ary_new();
     vm->loaded_features_snapshot = rb_ary_tmp_new(0);
     vm->loaded_features_index = st_init_numtable();
@@ -1197,9 +1314,9 @@ Init_load(void)
     rb_define_global_function("require", rb_f_require, 1);
     rb_define_global_function("require_relative", rb_f_require_relative, 1);
     rb_define_method(rb_cModule, "autoload", rb_mod_autoload, 2);
-    rb_define_method(rb_cModule, "autoload?", rb_mod_autoload_p, 1);
+    rb_define_method(rb_cModule, "autoload?", rb_mod_autoload_p, -1);
     rb_define_global_function("autoload", rb_f_autoload, 2);
-    rb_define_global_function("autoload?", rb_f_autoload_p, 1);
+    rb_define_global_function("autoload?", rb_f_autoload_p, -1);
 
     ruby_dln_librefs = rb_ary_tmp_new(0);
     rb_gc_register_mark_object(ruby_dln_librefs);
