@@ -5,30 +5,39 @@
   which is written in ruby.
 */
 
-#include "internal.h"
-#include "id.h"
-#include <math.h>
+#include "ruby/internal/config.h"
+
+#include <ctype.h>
 #include <float.h>
+#include <math.h>
 
 #ifdef HAVE_IEEEFP_H
 #include <ieeefp.h>
 #endif
 
-#define NDEBUG
-#include "ruby_assert.h"
-
 #if defined(HAVE_LIBGMP) && defined(HAVE_GMP_H)
 #define USE_GMP
 #include <gmp.h>
 #endif
 
+#undef NDEBUG
+#define NDEBUG
+#include "id.h"
+#include "internal.h"
+#include "internal/array.h"
+#include "internal/complex.h"
+#include "internal/gc.h"
+#include "internal/numeric.h"
+#include "internal/object.h"
+#include "internal/rational.h"
+#include "ruby_assert.h"
+
 #define ZERO INT2FIX(0)
 #define ONE INT2FIX(1)
 #define TWO INT2FIX(2)
 
 #define GMP_GCD_DIGITS 1
 
-#define INT_POSITIVE_P(x) (FIXNUM_P(x) ? FIXNUM_POSITIVE_P(x) : BIGNUM_POSITIVE_P(x))
 #define INT_ZERO_P(x) (FIXNUM_P(x) ? FIXNUM_ZERO_P(x) : rb_bigzero_p(x))
 
 VALUE rb_cRational;
@@ -52,6 +61,8 @@ f_add(VALUE x, VALUE y)
 	return x;
     if (FIXNUM_ZERO_P(x))
 	return y;
+    if (RB_INTEGER_TYPE_P(x))
+        return rb_int_plus(x, y);
     return rb_funcall(x, '+', 1, y);
 }
 
@@ -70,6 +81,10 @@ f_lt_p(VALUE x, VALUE y)
 {
     if (FIXNUM_P(x) && FIXNUM_P(y))
 	return (SIGNED_VALUE)x < (SIGNED_VALUE)y;
+    if (RB_INTEGER_TYPE_P(x)) {
+        VALUE r = rb_int_cmp(x, y);
+        if (!NIL_P(r)) return rb_int_negative_p(r);
+    }
     return RTEST(rb_funcall(x, '<', 1, y));
 }
 
@@ -129,11 +144,13 @@ f_to_i(VALUE x)
     return rb_funcall(x, id_to_i, 0);
 }
 
-inline static VALUE
+inline static int
 f_eqeq_p(VALUE x, VALUE y)
 {
     if (FIXNUM_P(x) && FIXNUM_P(y))
 	return x == y;
+    if (RB_INTEGER_TYPE_P(x))
+        return RTEST(rb_int_equal(x, y));
     return (int)rb_equal(x, y);
 }
 
@@ -394,9 +411,9 @@ nurat_s_new_internal(VALUE klass, VALUE num, VALUE den)
 {
     NEWOBJ_OF(obj, struct RRational, klass, T_RATIONAL | (RGENGC_WB_PROTECTED_RATIONAL ? FL_WB_PROTECTED : 0));
 
-    RRATIONAL_SET_NUM(obj, num);
-    RRATIONAL_SET_DEN(obj, den);
-    OBJ_FREEZE_RAW(obj);
+    RATIONAL_SET_NUM((VALUE)obj, num);
+    RATIONAL_SET_DEN((VALUE)obj, den);
+    OBJ_FREEZE_RAW((VALUE)obj);
 
     return (VALUE)obj;
 }
@@ -413,18 +430,6 @@ f_rational_new_bang1(VALUE klass, VALUE x)
     return nurat_s_new_internal(klass, x, ONE);
 }
 
-#ifdef CANONICALIZATION_FOR_MATHN
-static int canonicalization = 0;
-
-RUBY_FUNC_EXPORTED void
-nurat_canonicalization(int f)
-{
-    canonicalization = f;
-}
-#else
-# define canonicalization 0
-#endif
-
 inline static void
 nurat_int_check(VALUE num)
 {
@@ -473,8 +478,6 @@ nurat_s_canonicalize_internal(VALUE klass, VALUE num, VALUE den)
     nurat_canonicalize(&num, &den);
     nurat_reduce(&num, &den);
 
-    if (canonicalization && f_one_p(den))
-	return num;
     return nurat_s_new_internal(klass, num, den);
 }
 
@@ -483,30 +486,9 @@ nurat_s_canonicalize_internal_no_reduce(VALUE klass, VALUE num, VALUE den)
 {
     nurat_canonicalize(&num, &den);
 
-    if (canonicalization && f_one_p(den))
-	return num;
     return nurat_s_new_internal(klass, num, den);
 }
 
-static VALUE
-nurat_s_new(int argc, VALUE *argv, VALUE klass)
-{
-    VALUE num, den;
-
-    switch (rb_scan_args(argc, argv, "11", &num, &den)) {
-      case 1:
-	num = nurat_int_value(num);
-	den = ONE;
-	break;
-      default:
-	num = nurat_int_value(num);
-	den = nurat_int_value(den);
-	break;
-    }
-
-    return nurat_s_canonicalize_internal(klass, num, den);
-}
-
 inline static VALUE
 f_rational_new2(VALUE klass, VALUE x, VALUE y)
 {
@@ -774,8 +756,8 @@ rb_rational_plus(VALUE self, VALUE other)
  *    Rational(9, 8)  - 4                #=> (-23/8)
  *    Rational(20, 9) - 9.8              #=> -7.577777777777778
  */
-static VALUE
-nurat_sub(VALUE self, VALUE other)
+VALUE
+rb_rational_minus(VALUE self, VALUE other)
 {
     if (RB_INTEGER_TYPE_P(other)) {
 	{
@@ -976,15 +958,6 @@ nurat_fdiv(VALUE self, VALUE other)
     return rb_funcall(div, idTo_f, 0);
 }
 
-inline static VALUE
-f_odd_p(VALUE integer)
-{
-    if (rb_funcall(integer, '%', 1, INT2FIX(2)) != INT2FIX(0)) {
-	return Qtrue;
-    }
-    return Qfalse;
-}
-
 /*
  * call-seq:
  *    rat ** numeric  ->  numeric
@@ -1019,7 +992,7 @@ rb_rational_pow(VALUE self, VALUE other)
 		return f_rational_new_bang1(CLASS_OF(self), ONE);
 	    }
 	    else if (f_minus_one_p(dat->num) && RB_INTEGER_TYPE_P(other)) {
-		return f_rational_new_bang1(CLASS_OF(self), INT2FIX(f_odd_p(other) ? -1 : 1));
+		return f_rational_new_bang1(CLASS_OF(self), INT2FIX(rb_int_odd_p(other) ? -1 : 1));
 	    }
 	    else if (INT_ZERO_P(dat->num)) {
 		if (rb_num_negative_p(other)) {
@@ -1096,21 +1069,19 @@ rb_rational_pow(VALUE self, VALUE other)
 VALUE
 rb_rational_cmp(VALUE self, VALUE other)
 {
-    if (RB_INTEGER_TYPE_P(other)) {
+    switch (TYPE(other)) {
+      case T_FIXNUM:
+      case T_BIGNUM:
 	{
 	    get_dat1(self);
 
 	    if (dat->den == LONG2FIX(1))
 		return rb_int_cmp(dat->num, other); /* c14n */
 	    other = f_rational_new_bang1(CLASS_OF(self), other);
-	    goto other_is_rational;
+            /* FALLTHROUGH */
 	}
-    }
-    else if (RB_FLOAT_TYPE_P(other)) {
-	return rb_dbl_cmp(nurat_to_double(self), RFLOAT_VALUE(other));
-    }
-    else if (RB_TYPE_P(other, T_RATIONAL)) {
-	other_is_rational:
+
+      case T_RATIONAL:
 	{
 	    VALUE num1, num2;
 
@@ -1127,8 +1098,11 @@ rb_rational_cmp(VALUE self, VALUE other)
 	    }
 	    return rb_int_cmp(rb_int_minus(num1, num2), ZERO);
 	}
-    }
-    else {
+
+      case T_FLOAT:
+        return rb_dbl_cmp(nurat_to_double(self), RFLOAT_VALUE(other));
+
+      default:
 	return rb_num_coerce_cmp(self, other, rb_intern("<=>"));
     }
 }
@@ -1723,21 +1697,30 @@ static VALUE
 nurat_rationalize(int argc, VALUE *argv, VALUE self)
 {
     VALUE e, a, b, p, q;
+    VALUE rat = self;
+    get_dat1(self);
 
     if (rb_check_arity(argc, 0, 1) == 0)
 	return self;
 
-    if (nurat_negative_p(self))
-	return rb_rational_uminus(nurat_rationalize(argc, argv, rb_rational_uminus(self)));
-
     e = f_abs(argv[0]);
-    a = f_sub(self, e);
-    b = f_add(self, e);
+
+    if (INT_NEGATIVE_P(dat->num)) {
+        rat = f_rational_new2(RBASIC_CLASS(self), rb_int_uminus(dat->num), dat->den);
+    }
+
+    a = FIXNUM_ZERO_P(e) ? rat : rb_rational_minus(rat, e);
+    b = FIXNUM_ZERO_P(e) ? rat : rb_rational_plus(rat, e);
 
     if (f_eqeq_p(a, b))
 	return self;
 
     nurat_rationalize_internal(a, b, &p, &q);
+    if (rat != self) {
+        RATIONAL_SET_NUM(rat, rb_int_uminus(p));
+        RATIONAL_SET_DEN(rat, q);
+        return rat;
+    }
     return f_rational_new2(CLASS_OF(self), p, q);
 }
 
@@ -1827,8 +1810,8 @@ nurat_loader(VALUE self, VALUE a)
     nurat_int_check(num);
     nurat_int_check(den);
     nurat_canonicalize(&num, &den);
-    RRATIONAL_SET_NUM(dat, num);
-    RRATIONAL_SET_DEN(dat, den);
+    RATIONAL_SET_NUM((VALUE)dat, num);
+    RATIONAL_SET_DEN((VALUE)dat, den);
     OBJ_FREEZE_RAW(self);
 
     return self;
@@ -1853,7 +1836,6 @@ nurat_marshal_load(VALUE self, VALUE a)
     VALUE num, den;
 
     rb_check_frozen(self);
-    rb_check_trusted(self);
 
     Check_Type(a, T_ARRAY);
     if (RARRAY_LEN(a) != 2)
@@ -1939,6 +1921,14 @@ rb_gcdlcm(VALUE self, VALUE other)
 VALUE
 rb_rational_raw(VALUE x, VALUE y)
 {
+    if (! RB_INTEGER_TYPE_P(x))
+        x = rb_to_int(x);
+    if (! RB_INTEGER_TYPE_P(y))
+        y = rb_to_int(y);
+    if (INT_NEGATIVE_P(y)) {
+        x = rb_int_uminus(x);
+        y = rb_int_uminus(y);
+    }
     return nurat_s_new_internal(rb_cRational, x, y);
 }
 
@@ -2022,12 +2012,7 @@ rb_numeric_quo(VALUE x, VALUE y)
         return rb_funcallv(x, idFdiv, 1, &y);
     }
 
-    if (canonicalization) {
-        x = rb_rational_raw1(x);
-    }
-    else {
-        x = rb_convert_type(x, T_RATIONAL, "Rational", "to_r");
-    }
+    x = rb_convert_type(x, T_RATIONAL, "Rational", "to_r");
     return nurat_div(x, y);
 }
 
@@ -2086,9 +2071,6 @@ rb_float_numerator(VALUE self)
     if (isinf(d) || isnan(d))
 	return self;
     r = float_to_r(self);
-    if (canonicalization && k_integer_p(r)) {
-	return r;
-    }
     return nurat_numerator(r);
 }
 
@@ -2109,9 +2091,6 @@ rb_float_denominator(VALUE self)
     if (isinf(d) || isnan(d))
 	return INT2FIX(1);
     r = float_to_r(self);
-    if (canonicalization && k_integer_p(r)) {
-	return ONE;
-    }
     return nurat_denominator(r);
 }
 
@@ -2242,7 +2221,7 @@ rb_flt_rationalize_with_prec(VALUE flt, VALUE prec)
 VALUE
 rb_flt_rationalize(VALUE flt)
 {
-    VALUE a, b, f, p, q;
+    VALUE a, b, f, p, q, den;
     int n;
 
     float_decode_internal(flt, &f, &n);
@@ -2250,7 +2229,7 @@ rb_flt_rationalize(VALUE flt)
         return rb_rational_new1(rb_int_lshift(f, INT2FIX(n)));
 
     {
-        VALUE radix_times_f, den;
+        VALUE radix_times_f;
 
         radix_times_f = rb_int_mul(INT2FIX(FLT_RADIX), f);
 #if FLT_RADIX == 2 && 0
@@ -2259,13 +2238,15 @@ rb_flt_rationalize(VALUE flt)
         den = rb_int_positive_pow(FLT_RADIX, 1-n);
 #endif
 
-        a = rb_rational_new2(rb_int_minus(radix_times_f, INT2FIX(FLT_RADIX - 1)), den);
-        b = rb_rational_new2(rb_int_plus(radix_times_f, INT2FIX(FLT_RADIX - 1)), den);
+        a = rb_int_minus(radix_times_f, INT2FIX(FLT_RADIX - 1));
+        b = rb_int_plus(radix_times_f, INT2FIX(FLT_RADIX - 1));
     }
 
-    if (nurat_eqeq_p(a, b))
+    if (f_eqeq_p(a, b))
         return float_to_r(flt);
 
+    a = rb_rational_new2(a, den);
+    b = rb_rational_new2(b, den);
     nurat_rationalize_internal(a, b, &p, &q);
     return rb_rational_new2(p, q);
 }
@@ -2288,20 +2269,20 @@ static VALUE
 float_rationalize(int argc, VALUE *argv, VALUE self)
 {
     double d = RFLOAT_VALUE(self);
-
-    if (d < 0.0)
-        return rb_rational_uminus(float_rationalize(argc, argv, DBL2NUM(-d)));
+    VALUE rat;
+    int neg = d < 0.0;
+    if (neg) self = DBL2NUM(-d);
 
     if (rb_check_arity(argc, 0, 1)) {
-        return rb_flt_rationalize_with_prec(self, argv[0]);
+        rat = rb_flt_rationalize_with_prec(self, argv[0]);
     }
     else {
-        return rb_flt_rationalize(self);
+        rat = rb_flt_rationalize(self);
     }
+    if (neg) RATIONAL_SET_NUM(rat, rb_int_uminus(RRATIONAL(rat)->num));
+    return rat;
 }
 
-#include <ctype.h>
-
 inline static int
 issign(int c)
 {
@@ -2417,7 +2398,7 @@ parse_rat(const char *s, const char *const e, int strict, int raise)
 
     if (!read_num(&s, e, &num, &nexp)) {
 	if (strict) return Qnil;
-	return canonicalization ? ZERO : nurat_s_alloc(rb_cRational);
+	return nurat_s_alloc(rb_cRational);
     }
     den = ONE;
     if (s < e && *s == '/') {
@@ -2445,24 +2426,27 @@ parse_rat(const char *s, const char *const e, int strict, int raise)
     if (nexp != ZERO) {
         if (INT_NEGATIVE_P(nexp)) {
             VALUE mul;
-            if (!FIXNUM_P(nexp)) {
-              overflow:
-                return sign == '-' ? DBL2NUM(-HUGE_VAL) : DBL2NUM(HUGE_VAL);
+            if (FIXNUM_P(nexp)) {
+                mul = f_expt10(LONG2NUM(-FIX2LONG(nexp)));
+                if (! RB_FLOAT_TYPE_P(mul)) {
+                    num = rb_int_mul(num, mul);
+                    goto reduce;
+                }
             }
-            mul = f_expt10(LONG2NUM(-FIX2LONG(nexp)));
-            if (RB_FLOAT_TYPE_P(mul)) goto overflow;
-            num = rb_int_mul(num, mul);
+            return sign == '-' ? DBL2NUM(-HUGE_VAL) : DBL2NUM(HUGE_VAL);
         }
         else {
             VALUE div;
-            if (!FIXNUM_P(nexp)) {
-              underflow:
-                return sign == '-' ? DBL2NUM(-0.0) : DBL2NUM(+0.0);
+            if (FIXNUM_P(nexp)) {
+                div = f_expt10(nexp);
+                if (! RB_FLOAT_TYPE_P(div)) {
+                    den = rb_int_mul(den, div);
+                    goto reduce;
+                }
             }
-            div = f_expt10(nexp);
-            if (RB_FLOAT_TYPE_P(div)) goto underflow;
-            den = rb_int_mul(den, div);
+            return sign == '-' ? DBL2NUM(-0.0) : DBL2NUM(+0.0);
         }
+      reduce:
         nurat_reduce(&num, &den);
     }
 
@@ -2470,9 +2454,7 @@ parse_rat(const char *s, const char *const e, int strict, int raise)
 	num = negate_num(num);
     }
 
-    if (!canonicalization || den != ONE)
-	num = rb_rational_raw(num, den);
-    return num;
+    return rb_rational_raw(num, den);
 }
 
 static VALUE
@@ -2562,6 +2544,8 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
     VALUE a1 = numv, a2 = denv;
     int state;
 
+    assert(a1 != Qundef);
+
     if (NIL_P(a1) || NIL_P(a2)) {
         if (!raise) return Qnil;
         rb_raise(rb_eTypeError, "can't convert nil into Rational");
@@ -2577,21 +2561,47 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
             a2 = RCOMPLEX(a2)->real;
     }
 
-    if (RB_FLOAT_TYPE_P(a1)) {
+    if (RB_INTEGER_TYPE_P(a1)) {
+        // nothing to do
+    }
+    else if (RB_FLOAT_TYPE_P(a1)) {
         a1 = float_to_r(a1);
     }
+    else if (RB_TYPE_P(a1, T_RATIONAL)) {
+        // nothing to do
+    }
     else if (RB_TYPE_P(a1, T_STRING)) {
         a1 = string_to_r_strict(a1, raise);
         if (!raise && NIL_P(a1)) return Qnil;
     }
+    else if (!rb_respond_to(a1, idTo_r)) {
+        VALUE tmp = rb_protect(rb_check_to_int, a1, NULL);
+        rb_set_errinfo(Qnil);
+        if (!NIL_P(tmp)) {
+            a1 = tmp;
+        }
+    }
 
-    if (RB_FLOAT_TYPE_P(a2)) {
+    if (RB_INTEGER_TYPE_P(a2)) {
+        // nothing to do
+    }
+    else if (RB_FLOAT_TYPE_P(a2)) {
         a2 = float_to_r(a2);
     }
+    else if (RB_TYPE_P(a2, T_RATIONAL)) {
+        // nothing to do
+    }
     else if (RB_TYPE_P(a2, T_STRING)) {
         a2 = string_to_r_strict(a2, raise);
         if (!raise && NIL_P(a2)) return Qnil;
     }
+    else if (a2 != Qundef && !rb_respond_to(a2, idTo_r)) {
+        VALUE tmp = rb_protect(rb_check_to_int, a2, NULL);
+        rb_set_errinfo(Qnil);
+        if (!NIL_P(tmp)) {
+            a2 = tmp;
+        }
+    }
 
     if (RB_TYPE_P(a1, T_RATIONAL)) {
         if (a2 == Qundef || (k_exact_one_p(a2)))
@@ -2599,7 +2609,7 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
     }
 
     if (a2 == Qundef) {
-        if (!k_integer_p(a1)) {
+        if (!RB_INTEGER_TYPE_P(a1)) {
             if (!raise) {
                 VALUE result = rb_protect(to_rational, a1, NULL);
                 rb_set_errinfo(Qnil);
@@ -2634,25 +2644,32 @@ nurat_convert(VALUE klass, VALUE numv, VALUE denv, int raise)
             }
         }
         if ((k_numeric_p(a1) && k_numeric_p(a2)) &&
-                (!f_integer_p(a1) || !f_integer_p(a2)))
+                (!f_integer_p(a1) || !f_integer_p(a2))) {
+            VALUE tmp = rb_protect(to_rational, a1, &state);
+            if (!state) {
+                a1 = tmp;
+            }
+            else {
+                rb_set_errinfo(Qnil);
+            }
             return f_div(a1, a2);
+        }
     }
 
-    {
-        int argc;
-        VALUE argv2[2];
-        argv2[0] = a1;
-        if (a2 == Qundef) {
-            argv2[1] = Qnil;
-            argc = 1;
-        }
-        else {
-            if (!k_integer_p(a2) && !raise) return Qnil;
-            argv2[1] = a2;
-            argc = 2;
-        }
-        return nurat_s_new(argc, argv2, klass);
+    a1 = nurat_int_value(a1);
+
+    if (a2 == Qundef) {
+        a2 = ONE;
     }
+    else if (!k_integer_p(a2) && !raise) {
+        return Qnil;
+    }
+    else {
+        a2 = nurat_int_value(a2);
+    }
+
+
+    return nurat_s_canonicalize_internal(klass, a1, a2);
 }
 
 static VALUE
@@ -2713,13 +2730,10 @@ void
 Init_Rational(void)
 {
     VALUE compat;
-#undef rb_intern
-#define rb_intern(str) rb_intern_const(str)
-
-    id_abs = rb_intern("abs");
-    id_integer_p = rb_intern("integer?");
-    id_i_num = rb_intern("@numerator");
-    id_i_den = rb_intern("@denominator");
+    id_abs = rb_intern_const("abs");
+    id_integer_p = rb_intern_const("integer?");
+    id_i_num = rb_intern_const("@numerator");
+    id_i_den = rb_intern_const("@denominator");
 
     rb_cRational = rb_define_class("Rational", rb_cNumeric);
 
@@ -2735,7 +2749,7 @@ Init_Rational(void)
 
     rb_define_method(rb_cRational, "-@", rb_rational_uminus, 0);
     rb_define_method(rb_cRational, "+", rb_rational_plus, 1);
-    rb_define_method(rb_cRational, "-", nurat_sub, 1);
+    rb_define_method(rb_cRational, "-", rb_rational_minus, 1);
     rb_define_method(rb_cRational, "*", rb_rational_mul, 1);
     rb_define_method(rb_cRational, "/", nurat_div, 1);
     rb_define_method(rb_cRational, "quo", nurat_div, 1);
