@@ -53,10 +53,14 @@ class TestProc < Test::Unit::TestCase
     assert_equal(5, x)
   end
 
-  def assert_arity(n)
+  def assert_arity(n, &block)
     meta = class << self; self; end
-    meta.class_eval {define_method(:foo, Proc.new)}
-    assert_equal(n, method(:foo).arity)
+    b = Proc.new(&block)
+    meta.class_eval {
+      remove_method(:foo_arity) if method_defined?(:foo_arity)
+      define_method(:foo_arity, b)
+    }
+    assert_equal(n, method(:foo_arity).arity)
   end
 
   def test_arity
@@ -133,6 +137,14 @@ class TestProc < Test::Unit::TestCase
     lambda { x }
   end
 
+  def m_nest0(&block)
+    block
+  end
+
+  def m_nest(&block)
+    [m_nest0(&block), m_nest0(&block)]
+  end
+
   def test_eq
     a = m(1)
     b = m(2)
@@ -144,6 +156,8 @@ class TestProc < Test::Unit::TestCase
     a = lambda {|x| lambda {} }.call(1)
     b = lambda {}
     assert_not_equal(a, b, "[ruby-dev:22601]")
+
+    assert_equal(*m_nest{}, "[ruby-core:84583] Feature #14627")
   end
 
   def test_block_par
@@ -151,45 +165,6 @@ class TestProc < Test::Unit::TestCase
     assert_equal(12, Proc.new{|a,&b| b.call(a)}.call(12) {|x| x})
   end
 
-  def test_safe
-    safe = $SAFE
-    c = Class.new
-    x = c.new
-
-    p = proc {
-      $SAFE += 1
-      proc {$SAFE}
-    }.call
-
-    assert_equal(safe + 1, $SAFE)
-    assert_equal(safe + 1, p.call)
-    assert_equal(safe + 1, $SAFE)
-
-    $SAFE = 0
-    c.class_eval {define_method(:safe, p)}
-    assert_equal(safe, x.safe)
-
-    $SAFE = 0
-    p = proc {$SAFE += 1}
-    assert_equal(safe + 1, p.call)
-    assert_equal(safe + 1, $SAFE)
-
-    $SAFE = 0
-    c.class_eval {define_method(:inc, p)}
-    assert_equal(safe + 1, proc {x.inc; $SAFE}.call)
-    assert_equal(safe + 1, $SAFE)
-
-    $SAFE = 0
-    assert_equal(safe + 1, proc {x.method(:inc).call; $SAFE}.call)
-    assert_equal(safe + 1, $SAFE)
-
-    $SAFE = 0
-    assert_equal(safe + 1, proc {x.method(:inc).to_proc.call; $SAFE}.call)
-    assert_equal(safe + 1, $SAFE)
-  ensure
-    $SAFE = 0
-  end
-
   def m2
     "OK"
   end
@@ -817,6 +792,33 @@ class TestProc < Test::Unit::TestCase
     assert_equal [[1, 2], Proc, :x], (pr.call(1, 2){|x| x})
   end
 
+  def test_proc_args_only_rest
+    pr = proc {|*c| c }
+    assert_equal [], pr.call()
+    assert_equal [1], pr.call(1)
+    assert_equal [[1]], pr.call([1])
+    assert_equal [1, 2], pr.call(1,2)
+    assert_equal [[1, 2]], pr.call([1,2])
+  end
+
+  def test_proc_args_rest_kw
+    pr = proc {|*c, a: 1| [c, a] }
+    assert_equal [[], 1], pr.call()
+    assert_equal [[1], 1], pr.call(1)
+    assert_equal [[[1]], 1], pr.call([1])
+    assert_equal [[1, 2], 1], pr.call(1,2)
+    assert_equal [[[1, 2]], 1], pr.call([1,2])
+  end
+
+  def test_proc_args_rest_kwsplat
+    pr = proc {|*c, **kw| [c, kw] }
+    assert_equal [[], {}], pr.call()
+    assert_equal [[1], {}], pr.call(1)
+    assert_equal [[[1]], {}], pr.call([1])
+    assert_equal [[1, 2], {}], pr.call(1,2)
+    assert_equal [[[1, 2]], {}], pr.call([1,2])
+  end
+
   def test_proc_args_pos_rest_post_block
     pr = proc {|a,b,*c,d,e,&f|
       [a, b, c, d, e, f.class, f&&f.call(:x)]
@@ -1120,6 +1122,36 @@ class TestProc < Test::Unit::TestCase
     assert_equal([1,2,[3],4,5], r, "[ruby-core:19485]")
   end
 
+  def test_proc_autosplat
+    def self.a(arg, kw)
+      yield arg
+      yield arg, **kw
+      yield arg, kw
+    end
+
+    arr = []
+    a([1,2,3], {}) do |arg1, arg2=0|
+      arr << [arg1, arg2]
+    end
+    assert_equal([[1, 2], [[1, 2, 3], 0], [[1, 2, 3], {}]], arr)
+
+    arr = []
+    a([1,2,3], a: 1) do |arg1, arg2=0|
+      arr << [arg1, arg2]
+    end
+    assert_equal([[1, 2], [[1, 2, 3], {a: 1}], [[1, 2, 3], {a: 1}]], arr)
+  end
+
+  def test_proc_single_arg_with_keywords_accepted_and_yielded
+    def self.a
+      yield [], **{a: 1}
+    end
+    res = a do |arg, **opts|
+      [arg, opts]
+    end
+    assert_equal([[], {a: 1}], res)
+  end
+
   def test_parameters
     assert_equal([], proc {}.parameters)
     assert_equal([], proc {||}.parameters)
@@ -1191,9 +1223,6 @@ class TestProc < Test::Unit::TestCase
     assert_match(/^#<Proc:0x\h+ #{ Regexp.quote(__FILE__) }:\d+>$/, proc {}.to_s)
     assert_match(/^#<Proc:0x\h+ #{ Regexp.quote(__FILE__) }:\d+ \(lambda\)>$/, lambda {}.to_s)
     assert_match(/^#<Proc:0x\h+ \(lambda\)>$/, method(:p).to_proc.to_s)
-    x = proc {}
-    x.taint
-    assert_predicate(x.to_s, :tainted?)
     name = "Proc\u{1f37b}"
     assert_include(EnvUtil.labeled_class(name, Proc).new {}.to_s, name)
   end
@@ -1412,14 +1441,6 @@ class TestProc < Test::Unit::TestCase
     end;
   end
 
-  def method_for_test_proc_without_block_for_symbol
-    binding.eval('proc')
-  end
-
-  def test_proc_without_block_for_symbol
-    assert_equal('1', method_for_test_proc_without_block_for_symbol(&:to_s).call(1), '[Bug #14782]')
-  end
-
   def test_compose
     f = proc {|x| x * 2}
     g = proc {|x| x + 1}
@@ -1447,9 +1468,13 @@ class TestProc < Test::Unit::TestCase
   def test_compose_with_lambda
     f = lambda {|x| x * 2}
     g = lambda {|x| x}
+    not_lambda = proc {|x| x}
 
     assert_predicate((f << g), :lambda?)
     assert_predicate((g >> f), :lambda?)
+    assert_predicate((not_lambda << f), :lambda?)
+    assert_not_predicate((f << not_lambda), :lambda?)
+    assert_not_predicate((not_lambda >> f), :lambda?)
   end
 
   def test_compose_with_method
@@ -1461,6 +1486,7 @@ class TestProc < Test::Unit::TestCase
 
     assert_equal(6, (f << g).call(2))
     assert_equal(5, (f >> g).call(2))
+    assert_predicate((f << g), :lambda?)
   end
 
   def test_compose_with_callable
@@ -1472,6 +1498,7 @@ class TestProc < Test::Unit::TestCase
 
     assert_equal(6, (f << g).call(2))
     assert_equal(5, (f >> g).call(2))
+    assert_predicate((f << g), :lambda?)
   end
 
   def test_compose_with_noncallable
@@ -1511,6 +1538,57 @@ class TestProc < Test::Unit::TestCase
     assert_equal(42, Module.new { extend self
       def m1(&b) b end; def m2(); m1 { next 42 } end }.m2.call)
   end
+
+  def test_isolate
+    assert_raise_with_message ArgumentError, /\(a\)/ do
+      a = :a
+      Proc.new{p a}.isolate.call
+    end
+
+    assert_raise_with_message ArgumentError, /\(a\)/ do
+      a = :a
+      1.times{
+        Proc.new{p a}.isolate.call
+      }
+    end
+
+    assert_raise_with_message ArgumentError, /yield/ do
+      Proc.new{yield}.isolate.call
+    end
+
+    # binding
+
+    :a.tap{|a|
+      :b.tap{|b|
+        Proc.new{
+          :c.tap{|c|
+            assert_equal :c, eval('c')
+
+            assert_raise_with_message SyntaxError, /\`a\'/ do
+              eval('p a')
+            end
+
+            assert_raise_with_message SyntaxError, /\`b\'/ do
+              eval('p b')
+            end
+
+            assert_raise_with_message SyntaxError, /can not yield from isolated Proc/ do
+              eval('p yield')
+            end
+
+            assert_equal :c, binding.local_variable_get(:c)
+
+            assert_raise_with_message NameError, /local variable \`a\' is not defined/ do
+              binding.local_variable_get(:a)
+            end
+
+            assert_equal [:c], local_variables
+            assert_equal [:c], binding.local_variables
+          }
+        }.isolate.call
+      }
+    }
+  end if proc{}.respond_to? :isolate
 end
 
 class TestProcKeywords < Test::Unit::TestCase
@@ -1519,29 +1597,15 @@ class TestProcKeywords < Test::Unit::TestCase
     g = ->(kw) { kw.merge(:a=>2) }
 
     assert_equal(2, (f >> g).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call(a: 3)[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call(a: 3)[:a] }
     assert_equal(2, (f >> g).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (f >> g).call({a: 3})[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f >> g).call({a: 3})[:a] }
     assert_equal(2, (g << f).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call(a: 3)[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (g << f).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call({a: 3})[:a])
-    end
-    assert_warn(/The keyword argument is passed as the last hash parameter.*The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call(**{})[:a])
-    end
+    assert_raise(ArgumentError) { (g >> f).call(a: 3)[:a] }
+    assert_raise(ArgumentError) { (g << f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (g >> f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f << g).call(**{})[:a] }
     assert_equal(2, (f >> g).call(**{})[:a])
   end
 
@@ -1549,29 +1613,15 @@ class TestProcKeywords < Test::Unit::TestCase
     f = ->(**kw) { kw.merge(:a=>1) }.method(:call)
     g = ->(kw) { kw.merge(:a=>2) }.method(:call)
 
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call(a: 3)[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call(a: 3)[:a] }
     assert_equal(2, (f >> g).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (f >> g).call({a: 3})[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f >> g).call({a: 3})[:a] }
     assert_equal(2, (g << f).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call(a: 3)[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (g << f).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call({a: 3})[:a])
-    end
-    assert_warn(/The keyword argument is passed as the last hash parameter.*The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call(**{})[:a])
-    end
+    assert_raise(ArgumentError) { (g >> f).call(a: 3)[:a] }
+    assert_raise(ArgumentError) { (g << f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (g >> f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f << g).call(**{})[:a] }
     assert_equal(2, (f >> g).call(**{})[:a])
   end
 
@@ -1583,29 +1633,15 @@ class TestProcKeywords < Test::Unit::TestCase
     def g.<<(f) to_proc << f end
     def g.>>(f) to_proc >> f end
 
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call(a: 3)[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call(a: 3)[:a] }
     assert_equal(2, (f >> g).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (f << g).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (f >> g).call({a: 3})[:a])
-    end
+    assert_raise(ArgumentError) { (f << g).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f >> g).call({a: 3})[:a] }
     assert_equal(2, (g << f).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call(a: 3)[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(2, (g << f).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for method/m) do
-      assert_equal(1, (g >> f).call({a: 3})[:a])
-    end
-    assert_warn(/The keyword argument is passed as the last hash parameter.*for `call'/m) do
-      assert_equal(1, (f << g).call(**{})[:a])
-    end
+    assert_raise(ArgumentError) { (g >> f).call(a: 3)[:a] }
+    assert_raise(ArgumentError) { (g << f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (g >> f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f << g).call(**{})[:a] }
     assert_equal(2, (f >> g).call(**{})[:a])
 
     f = ->(kw) { kw.merge(:a=>1) }
@@ -1616,29 +1652,15 @@ class TestProcKeywords < Test::Unit::TestCase
     def g.>>(f) to_proc >> f end
 
     assert_equal(1, (f << g).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(2, (f >> g).call(a: 3)[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(1, (f << g).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(2, (f >> g).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(2, (g << f).call(a: 3)[:a])
-    end
+    assert_raise(ArgumentError) { (f >> g).call(a: 3)[:a] }
+    assert_raise(ArgumentError) { (f << g).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (f >> g).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (g << f).call(a: 3)[:a] }
     assert_equal(1, (g >> f).call(a: 3)[:a])
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(2, (g << f).call({a: 3})[:a])
-    end
-    assert_warn(/The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(1, (g >> f).call({a: 3})[:a])
-    end
+    assert_raise(ArgumentError) { (g << f).call({a: 3})[:a] }
+    assert_raise(ArgumentError) { (g >> f).call({a: 3})[:a] }
     assert_equal(1, (f << g).call(**{})[:a])
-    assert_warn(/The keyword argument is passed as the last hash parameter.*The last argument is used as the keyword parameter.*for `call'/m) do
-      assert_equal(2, (f >> g).call(**{})[:a])
-    end
+    assert_raise(ArgumentError) { (f >> g).call(**{})[:a] }
   end
 end
 
