@@ -64,7 +64,18 @@ module Test
         args = @init_hook.call(args, options) if @init_hook
         non_options(args, options)
         @run_options = orig_args
-        @help = orig_args.map { |s| s =~ /[\s|&<>$()]/ ? s.inspect : s }.join " "
+
+        if seed = options[:seed]
+          srand(seed)
+        else
+          seed = options[:seed] = srand % 100_000
+          srand(seed)
+          orig_args.unshift "--seed=#{seed}"
+        end
+
+        @help = "\n" + orig_args.map { |s|
+          "  " + (s =~ /[\s|&<>$()]/ ? s.inspect : s)
+        }.join("\n")
         @options = options
       end
 
@@ -79,7 +90,7 @@ module Test
         end
 
         opts.on '-s', '--seed SEED', Integer, "Sets random seed" do |m|
-          options[:seed] = m
+          options[:seed] = m.to_i
         end
 
         opts.on '-v', '--verbose', "Verbose. Show progress processing files." do
@@ -91,7 +102,7 @@ module Test
           (options[:filter] ||= []) << a
         end
 
-        opts.on '--test-order=random|alpha|sorted', [:random, :alpha, :sorted] do |a|
+        opts.on '--test-order=random|alpha|sorted|nosort', [:random, :alpha, :sorted, :nosort] do |a|
           MiniTest::Unit::TestCase.test_order = a
         end
       end
@@ -114,6 +125,7 @@ module Test
             filter = /\A(?=.*#{filter})(?!.*#{negative})/
           end
           if Regexp === filter
+            filter = filter.dup
             # bypass conversion in minitest
             def filter.=~(other)    # :nodoc:
               super unless Regexp === other
@@ -191,6 +203,10 @@ module Test
         opts.on '--ruby VAL', "Path to ruby which is used at -j option" do |a|
           options[:ruby] = a.split(/ /).reject(&:empty?)
         end
+
+        opts.on '--timetable-data=FILE', "Path to timetable data" do |a|
+          options[:timetable_data] = a
+        end
       end
 
       class Worker
@@ -204,8 +220,12 @@ module Test
         end
 
         attr_reader :quit_called
+        attr_accessor :start_time
+
+        @@worker_number = 0
 
         def initialize(io, pid, status)
+          @num = (@@worker_number += 1)
           @io = io
           @pid = pid
           @status = status
@@ -216,6 +236,10 @@ module Test
           @quit_called = false
         end
 
+        def name
+          "Worker #{@num}"
+        end
+
         def puts(*args)
           @io.puts(*args)
         end
@@ -228,6 +252,7 @@ module Test
             @loadpath = $:.dup
             puts "run #{task} #{type}"
             @status = :prepare
+            @start_time = Time.now
           rescue Errno::EPIPE
             died
           rescue IOError
@@ -320,6 +345,19 @@ module Test
         warn "or, a bug of test/unit/parallel.rb. try again without -j"
         warn "option."
         warn ""
+        if File.exist?('core')
+          require 'fileutils'
+          require 'time'
+          Dir.glob('/tmp/test-unit-core.*').each do |f|
+            if Time.now - File.mtime(f) > 7 * 24 * 60 * 60 # 7 days
+              warn "Deleting an old core file: #{f}"
+              FileUtils.rm(f)
+            end
+          end
+          core_path = "/tmp/test-unit-core.#{Time.now.utc.iso8601}"
+          warn "A core file is found. Saving it at: #{core_path.dump}"
+          FileUtils.mv('core', core_path)
+        end
         STDERR.flush
         exit c
       end
@@ -393,6 +431,7 @@ module Test
         worker = @workers_hash[io]
         cmd = worker.read
         cmd.sub!(/\A\.+/, '') if cmd # read may return nil
+
         case cmd
         when ''
           # just only dots, ignore
@@ -425,10 +464,19 @@ module Test
           rep    << {file: worker.real_file, report: r[2], result: r[3], testcase: r[5]}
           $:.push(*r[4]).uniq!
           jobs_status(worker) if @options[:job_status] == :replace
+
           return true
         when /^record (.+?)$/
           begin
             r = Marshal.load($1.unpack("m")[0])
+
+            suite = r.first
+            key = [worker.name, suite]
+            if @records[key]
+              @records[key][1] = worker.start_time = Time.now
+            else
+              @records[key] = [worker.start_time, Time.now]
+            end
           rescue => e
             print "unknown record: #{e.message} #{$1.unpack("m")[0].dump}"
             return true
@@ -455,6 +503,8 @@ module Test
       end
 
       def _run_parallel suites, type, result
+        @records = {}
+
         if @options[:parallel] < 1
           warn "Error: parameter of -j option should be greater than 0."
           return
@@ -463,6 +513,16 @@ module Test
         # Require needed thing for parallel running
         require 'timeout'
         @tasks = @files.dup # Array of filenames.
+
+        case MiniTest::Unit::TestCase.test_order
+        when :random
+          @tasks.shuffle!
+        else
+          # JIT first
+          ts = @tasks.group_by{|e| /test_jit/ =~ e ? 0 : 1}
+          @tasks = ts[0] + ts[1] if ts.size == 2
+        end
+
         @need_quit = false
         @dead_workers = []  # Array of dead workers.
         @warnings = []
@@ -494,6 +554,14 @@ module Test
           @interrupt = ex
           return result
         ensure
+          if file = @options[:timetable_data]
+            open(file, 'w'){|f|
+              @records.each{|(worker, suite), (st, ed)|
+                f.puts '[' + [worker.dump, suite.dump, st.to_f * 1_000, ed.to_f * 1_000].join(", ") + '],'
+              }
+            }
+          end
+
           if @interrupt
             @ios.select!{|x| @workers_hash[x].status == :running }
             while !@ios.empty? && (__io = IO.select(@ios,[],[],10))
@@ -508,7 +576,7 @@ module Test
             parallel = @options[:parallel]
             @options[:parallel] = false
             suites, rep = rep.partition {|r| r[:testcase] && r[:file] && r[:report].any? {|e| !e[2].is_a?(MiniTest::Skip)}}
-            suites.map {|r| r[:file]}.uniq.each {|file| require file}
+            suites.map {|r| File.realpath(r[:file])}.uniq.each {|file| require file}
             suites.map! {|r| eval("::"+r[:testcase])}
             del_status_line or puts
             unless suites.empty?
@@ -922,13 +990,16 @@ module Test
       end
     end
 
-    module GCStressOption # :nodoc: all
+    module GCOption # :nodoc: all
       def setup_options(parser, options)
         super
         parser.separator "GC options:"
         parser.on '--[no-]gc-stress', 'Set GC.stress as true' do |flag|
           options[:gc_stress] = flag
         end
+        parser.on '--[no-]gc-compact', 'GC.compact every time' do |flag|
+          options[:gc_compact] = flag
+        end
       end
 
       def non_options(files, options)
@@ -945,6 +1016,18 @@ module Test
             end
           end
         end
+        if options.delete(:gc_compact)
+          MiniTest::Unit::TestCase.class_eval do
+            oldrun = instance_method(:run)
+            define_method(:run) do |runner|
+              begin
+                oldrun.bind_call(self, runner)
+              ensure
+                GC.compact
+              end
+            end
+          end
+        end
         super
       end
     end
@@ -1080,7 +1163,7 @@ module Test
       include Test::Unit::GlobOption
       include Test::Unit::RepeatOption
       include Test::Unit::LoadPathOption
-      include Test::Unit::GCStressOption
+      include Test::Unit::GCOption
       include Test::Unit::ExcludesOption
       include Test::Unit::TimeoutOption
       include Test::Unit::RunCount
